import logging
import requests
import json
import random
import datetime


log = logging.getLogger("domain_stats")

def dateconverter(o):
    if isinstance(o, datetime.datetime):
        return o.strftime("%Y-%m-%d %H:%M:%S")

class IscConnection():
    def __init__(self, enabled = True, login=None, token = None):
        self.isc_enabled = enabled
        self.isc_login = login
        self.isc_authtoken = token
        self.prohibited_tlds = []

    1#to isc
    #{"command" : "query",  "domain": toplevel.tld}   
    #response:
        ### ISC Response is a JSON packet in the following format:
        # { "seen_by_web" =  Domain Created Date in format YYYY-mm-DD HH:MM:SS,   (If domain has multiple dates this is the most recent)
        #   "expires" = domain expiration date Date in format YYYY-mm-DD HH:MM:SS,  (if domain has multiple dates this is the most recent)
        #    "seen_by_isc" = date when ISC first queried whoisxml to build this record in format YYYY-mm-DD HH:MM:SS,als
        #     "alerts"  =  A list of strings to make the user aware of regarding the domain in question (for later use) formta is ['alert1','alert2']
        ##STub an ISC response

    #When a client starts it sends a status to isc.  ISC can force database updates on client, deny client access
    #To isc
    #{"command": "status", "client_version":"client version number","database_version":database version info,  cache_efficiency": [hit,miss,expired], "database_efficiency": [hit,miss]}
    #response:
    # {"current_database_version": "database version expected", "interval": "number of seconds ISC wishes to wait until next status update", "deny_client": "client message"}

    def get_status(self,client_version, database_version, cache_stats, database_stats):
        #To isc
        #submit = json.dumps({"command":"status", "client_version":client_version, "database_version":database_version, "cache_efficiency":[cache.hits, cache.miss, cache.expire], "database_efficiency":[database_stats.hits, database_stats.miss])
        #{"command": "status", "client_version":"client version number","database_version":database version info,  cache_efficiency": [hit,miss,expired], "database_efficiency": [hit,miss]}
        #response from isc
        # { "interval": "number of minutes ISC wishes to wait until next status update", "deny_client": "client message"}
        #functions returns tuple:
        #       Position 1 of Tuple is boolean representing "CRITICAL ERROR" which, when True causes the program to abort.
        #       a list of messages to give to the clients regarding its health and operating ability
        fake_isc_response = json.dumps({"interval":5, "deny_client":""})
        isc_response = json.loads(fake_isc_response)
        interval = isc_response.get("interval", 60)
        deny_client = isc_response.get("deny_client")
        if deny_client:
            reason = ",".join(deny_client)
            log.info("The ISC denied the client. Reason: {}".format(reason))
            self.isc_enable = False
        return interval

    def get_config(self):
        #To isc
        #submit = json.dumps({"command":"config"})
        # 
        #response from isc
        # {"min_database_version": database version expected","min_client_version":client version expected,prohibited_tlds:[] }
        #functions returns tuple:
        #       Position 1 min client version
        #       Position 2 in min database version
        fake_isc_response = json.dumps({"min_database_version":1.2, "min_client_version":1.0, "prohibited_tlds": ['.local','.arpa]']})
        isc_response = json.loads(fake_isc_response)
        self.prohibited_tlds.extend(isc_response.get("prohibited_tlds", []))
        return isc_response.get("min_client_version"), isc_response.get("min_database_version")

    def health_check(self, client_version, database_version, cache, database_stats):
        #To isc
        #submit = json.dumps({"command":"status", "client_version":client_version, "database_version":database_version, "cache_efficiency":[cache.hits, cache.miss, cache.expire], "database_efficiency":[database_stats.hits, database_stats.miss])
        #{"command": "status", "client_version":"client version number","database_version":database version info,  cache_efficiency": [hit,miss,expired], "database_efficiency": [hit,miss]}
        #response from isc
        # {"expected_database_version": "database version expected","expected_client_version":client version expected,  "interval": "number of minutes ISC wishes to wait until next status update", "deny_client": "client message", "notice":['messages' 'to' 'client']}
        #functions returns tuple:
        #       Position 1 of Tuple is boolean representing "CRITICAL ERROR" which, when True causes the program to abort.
        #       a list of messages to give to the clients regarding its health and operating ability
        client_messages = []
        fake_isc_response = json.dumps({"expected_database_version":1.1, "expected_client_version":1.0, "interval":5, "deny_client":"", "notice":['message1']})
        isc_response = json.loads(fake_isc_response)
        expected_database = isc_response.get("expected_database_version")
        if expected_database != database_version:
            log.debug("Database is out of date. ISC expected {}. Running {}".format(expected_database, database_version) )
            return ( True, 0.1 , ["UPDATE-DATABASE", str(expected_database)] ) 
        expected_client = isc_response.get("expected_client_version")
        if expected_client != client_version:
            log.debug("Warning Client Software is out of data.")
            client_messages.append("Your version of domain_stats is out of date.  Please update.")
        deny_client = isc_response.get("deny_client")
        if deny_client:
            client_messages.insert(0, deny_client)
        client_messages.extend(isc_response.get("notice"))
        return ( deny_client, isc_response.get("interval", 60), client_messages )

    def get_server_config(self):
        #To isc
        #submit = json.dumps({"command":"config"}, "client_version":client_version, "database_version":database_version})
        # 
        #response from isc
        # {"expected_database_version": "database version expected","expected_client_version":client version expected, "prohibited_tlds":['.local' 'to' 'arpa']}
        #resp = request.post( , configrequest)
        resp = json.loads(json.dumps({'expected_database_version':1.0, 'expected_client_version':1.0, 'prohibited_tlds': ['.local', '.arpa'] }))
        if sorted(resp.keys()) != ['expected_client_version', 'expected_database_version', 'prohibited_tlds']:
            print("Invalid ISC response requesting server config {}".format(resp))
            log.info("Invalid ISC response requesting server config {0}".format(resp))
        return resp


    def retrieve_isc(self, domain):
        #to isc
        #{"command" : "query",  "domain": toplevel.tld} 
        ### ISC Response is a JSON packet in the following format:
        # { "seen_by_web" =  Domain Created Date in format YYYY-mm-DD HH:MM:SS,   (If domain has multiple dates this is the most recent)
        #   "expires" = domain expiration date Date in format YYYY-mm-DD HH:MM:SS,  (if domain has multiple dates this is the most recent)
        #    "seen_by_isc" = date when ISC first queried whoisxml to build this record in format YYYY-mm-DD HH:MM:SS,als
        #     "alerts"  =  A list of strings to make the user aware of regarding the domain in question (for later use) formta is ['alert1','alert2']
        ##Stub an ISC response
        if not self.isc_enabled:
            return ("ERROR","ERROR", 0, ['ISC Lookups disabled. See Log for details.'])
        fake_alerts = []
        fake_date1 = (datetime.datetime.now() - datetime.timedelta(days=random.randrange(365,3000))).replace(microsecond=0).isoformat().replace("T"," ")
        if bool(random.getrandbits(1)):
            fake_date1= datetime.datetime.utcnow()
            fake_alerts = ['ISC-FIRST-CONTACT']
        fake_date2 = (datetime.datetime.now() - datetime.timedelta(days=random.randrange(365,3000))).replace(microsecond=0).isoformat().replace("T"," ")
        fake_date3 = (datetime.datetime.now() + datetime.timedelta(days=random.randrange(365,3000))).replace(microsecond=0).isoformat().replace("T"," ")
        fake_isc_response1 = json.dumps({"seen_by_web":fake_date2, "expires":fake_date3, "seen_by_isc":fake_date1, "alerts":fake_alerts}, default=dateconverter)
        #ISC can also generate Error responses and define how long the error is cached on the client (preventing repeated queries)
        #Those responses look like this:
        #{ "seen_by_web" = "ERROR",
        #   "expires = "ERROR"
        #   "seen_by_isc = hours to live for this error in client cache for the queried domain. 0=dont cache,-1=dont expire but allow lru,-2=permenant (use with caution)"
        #   "alerts" = ['alert1','alert2' ]  list if alerts to associate with this query.  }
        fake_error_ttl = random.randrange(-1,5)
        alertoptions= ['bad thing happened', 'domain doesnt exist','login expired','isc temporarily unavailable', 'bad domain','bad record from registrar','unable to connect to whois']
        fake_alerts = [random.choice(alertoptions) for _ in range(random.randrange(1,3))]
        fake_isc_response2 = json.dumps({"seen_by_web":"ERROR", "expires":"ERROR", "seen_by_isc":fake_error_ttl, "alerts":fake_alerts}, default=dateconverter)
        fake_isc_response = random.choice([ fake_isc_response1]*5 + [fake_isc_response2])
        #Process ISC response
        resp = json.loads(fake_isc_response)
        if sorted(resp.keys()) != ['alerts', 'expires', 'seen_by_isc', 'seen_by_web']:
            log.info("INVALID ISC RESPONSE MISSING KEY FIELDS {0}".format(resp))
            raise Exception("ISC RESPONSE to domain request MISSING KEY FIELDS.")
        web = resp['seen_by_web']
        if web != "ERROR":
            web = datetime.datetime.strptime(web, '%Y-%m-%d %H:%M:%S')
            expire = datetime.datetime.strptime(resp['expires'], '%Y-%m-%d %H:%M:%S')
            isc = resp['seen_by_isc']
            if isc != "FIRST-CONTACT":
                isc = datetime.datetime.strptime(isc, '%Y-%m-%d %H:%M:%S')
            return (web,expire, isc, resp['alerts'])
        return ("ERROR","ERROR", resp['seen_by_isc'], resp['alerts'])


from diskcache import FanoutCache
import datetime

class ExpiringCache:

    def __init__(self, cache_path):
        self.cache = FanoutCache(directory=cache_path, timeout=2, retry=True)

    def set(self, domain, domain_record, seconds_to_live, tag=None): 
        self.cache.set(domain, domain_record, expire = seconds_to_live, tag=tag)

    def get(self, key):
        return self.cache.get(key)

    def cache_dump(self,offset=0, limit=100):
        res = []
        for eachkey in self.cache:
            if offset != 0:
                offset -= 1
                continue
            if limit == 0:
                break
            val, expires = self.cache.get(eachkey, expire_time=True)
            exp_date = datetime.datetime.fromtimestamp(expires)
            res.append( {"domain":eachkey,"cache_value": val.get_json() ,"cache_expires":exp_date} )
            limit -= 1
        return res

    def cache_get(self,key):
        val, expires = self.cache.get(key, expire_time=True)
        exp_date = datetime.datetime.fromtimestamp(expires)
        res = {"domain":key,"cache_value": val.get_json() ,"cache_expires":exp_date} 
        return res

    def cache_info(self):
        hits,miss = self.cache.stats()
        size = self.cache.volume()
        warnings = self.cache.check(fix=True)
        return {"hit":hits, "miss":miss, "size": size, "warnings":warnings}




import logging
import requests
import rdap
import json
import datetime
import dateutil.parser
import dateutil.tz
import pathlib
import os
import sqlite3

log = logging.getLogger("domain_stats")

def get_config(update_url):
    try:
        config = requests.get( update_url + "/version.json").json()
    except Exception as e:
        print("Unable to retreive /version.json info from {} {}".format( update_url,str(e)))
        return 9999,99999
    return config.get("min_client_version"), config.get("min_database_version")
    

def retrieve_data(action_name,eventlist):
    for entry in eventlist:
        if entry.get("eventAction") == action_name:
            return entry.get("eventDate")
    return None

def get_domain_record(domain):
    client = rdap.client.RdapClient()
    client.url = "https://www.rdap.net"
    client.timeout = 5
    log.debug("RDAP Query of domain {0}".format(domain)) 
    try:
        resp = client.get_domain(domain).data
    except Exception as e:
        return "ERROR","ERROR", str(e)
    reg = retrieve_data( 'registration', resp.get('events'))
    reg = dateutil.parser.isoparse(reg)
    reg.replace(tzinfo=datetime.timezone.utc)
    exp = retrieve_data( 'expiration', resp.get('events'))
    try:
        exp = dateutil.parser.isoparse(exp)
    except Exception as e:
        return "ERROR","ERROR", "Invalid RDAP response Domain:{} generated {}".format(domain,str(e))
    exp.replace(tzinfo=datetime.timezone.utc)
    log.debug("RDAP Query of domain {0} resolved {1}".format(domain,resp)) 
    return reg,exp,""


import pathlib
import os
import sys
from domain_stats.settings import setup_directory



def launch_from_config(tgt_folder):
    os.chdir(tgt_folder)
    launch_cmd = f"""{sys.executable} -m gunicorn.app.wsgiapp 'domain_stats.server:config_app("{tgt_folder}")' -c {tgt_folder}/gunicorn_config.py"""
    os.system(launch_cmd)

def main():
    if len(sys.argv) != 2:
        typed_folder = input("Where do you want to store your domain_stats data and binaries? ")
    else:
        typed_folder = sys.argv[1]

    tgt_folder = pathlib.Path(typed_folder)
    if not tgt_folder.is_dir():
        print("That directory does not exist. Please create it and/or try again.")
        sys.exit(1)
    if (tgt_folder / "domain_stats.yaml").is_file():
        print("Existing config found in directory. Using it.")
        launch_from_config(tgt_folder)
    elif input("Would you like to setup this directory now?").lower().startswith("y"):
        setup_directory(tgt_folder)


if __name__ == "__main__":
    main()




import pathlib
import os
import sys
import shutil
import yaml
import multiprocessing


def create_gunicorn_config(tgt_folder,ip,port,cpu,thread):
    gunicorn_config = f"""
import gunicorn
import multiprocessing
import os

os.environ["SERVER_SOFTWARE"] = "domain_stats"
bind = "{ip}:{port}"
workers = {cpu}
threads = {thread}
gunicorn.SERVER_SOFTWARE = 'domain_stats'
    """
    config_file = pathlib.Path(tgt_folder / "gunicorn_config.py")
    config_file.write_text(gunicorn_config)

def update_setting(config, key, default):
    current_val = config.get(key,default)
    new_val = input(f"Set value for {key}. Default={default} Current={current_val} (Enter to keep Current): ") or current_val
    config[key]= default.__class__(new_val)    

def setup_directory(tgt_folder):
    config = {}
    file_path = tgt_folder / "domain_stats.yaml"
    if file_path.is_file():
        print("Existing config found in directory. Using it.") 
        config = yaml.safe_load(file_path.read_text())
        if not config:
            print("domain_stats.yaml config exists but is blank or could not be loaded. Erase the file to begin again.")
            sys.exit(1)
    source_folder = pathlib.Path(os.path.realpath(__file__)).parent
    if not (tgt_folder / "freqtable2018.freq").is_file():
        shutil.copy(source_folder / "data/freqtable2018.freq", tgt_folder)
    #create gunicorn config
    update_setting(config, "ip_address", "0.0.0.0")
    update_setting(config, "local_port", "5730")
    rec_workers = multiprocessing.cpu_count() * 2 + 1
    update_setting(config, "workers", rec_workers)
    rec_threads = multiprocessing.cpu_count() * 3
    update_setting(config, "threads_per_worker", rec_threads)
    #Configure Additional settings in domain_stats.yaml
    update_setting(config,'timezone_offset', 0)
    update_setting(config,'established_days_age', 730)
    update_setting(config,'mode',"rdap")
    update_setting(config,"rdap_error_ttl_days",7)
    update_setting(config,'freq_table', 'freqtable2018.freq')
    update_setting(config,'enable_freq_scores', True)
    update_setting(config,'freq_avg_alert',5.0)
    update_setting(config,'freq_word_alert',4.0)
    update_setting(config,'log_detail',0)
    update_setting(config,'cache_browse_limit',100)
    if input("Commit Changes to disk?").lower().startswith("y"):
        create_gunicorn_config(tgt_folder, config['ip_address'], config['local_port'],config['workers'], config['threads_per_worker'])
        with file_path.open('w') as fp:
            yaml.dump(config, fp, default_flow_style=False)

def main():
    if len(sys.argv) != 2:
        typed_folder = input("Where do you want to store your domain_stats data and binaries? ")
    else:
        typed_folder = sys.argv[1]

    tgt_folder = pathlib.Path(typed_folder)
    if not tgt_folder.is_dir():
        print("That directory does not exist. Please create it and/or try again.")
        sys.exit(1)

    setup_directory(tgt_folder)

if __name__ == "__main__":
    main()


    


import logging
from flask import Flask, jsonify, request, session, abort
import functools
import datetime
import argparse
import pathlib
import sys
import os

log = logging.getLogger("domain_stats")
default_flask_logger = logging.getLogger('werkzeug')
default_flask_logger.setLevel(logging.CRITICAL)


from publicsuffixlist import PublicSuffixList
from domain_stats.config import Config
from domain_stats.network_io import IscConnection
import domain_stats.rdap_query as rdap
from domain_stats.freq import FreqCounter
from domain_stats.expiring_diskcache import ExpiringCache
from diskcache import Cache

#current directory must be set by launcher to the location of the config and database
cache = ExpiringCache(os.getcwd())
memocache = Cache(pathlib.Path().cwd() / "memocache")
if not memocache.get("rdap_good"):
    memocache.set("rdap_good",0)
    memocache.set("rdap_fail",0)
config = Config(os.getcwd()+"/domain_stats.yaml")

if config.get("enable_freq_scores"):
    freq = FreqCounter()
    freq.load(config.get("freq_table"))

logfile = logging.FileHandler(str(pathlib.Path.cwd() / 'domain_stats.log'))
logformat = logging.Formatter('%(asctime)s : %(levelname)s : %(module)s : %(process)d : %(message)s')
logfile.setFormatter(logformat)
if config['log_detail'] == 0:
    log.setLevel(level=logging.CRITICAL)
elif config['log_detail'] == 1:
    log.addHandler(logfile)
    log.setLevel(logging.INFO)
else:
    log.addHandler(logfile)
    log.setLevel(logging.DEBUG)

app = Flask(__name__)
app.name = "domain_stats"
app.config['SESSION_TYPE'] = 'filesystem'
app.config['SECRET_KEY'] ="Change This to a unique very long string for your installation. Make it a few hundred characters long. You never have to remember it or use it, but you want it to be unique to your install."
app.config['PERMANENT_SESSION_LIFETIME'] = datetime.timedelta(hours=10)
#app.logger.removeHandler(default_handler)

def json_response(web,isc,you,cat,alert,freq=None):
    resp = {"seen_by_web":web,"seen_by_isc":isc, "seen_by_you":you, "category":cat, "alerts":alert}
    if app.config.get("enable_freq_scores"):
        resp['freq_score'] = freq
    return jsonify(resp)

@memocache.memoize(tag="reduce_domain")
def reduce_domain(domain_in):
    if not PublicSuffixList().publicsuffix(domain_in,accept_unknown=False):
        return None
    domain = PublicSuffixList().privatesuffix(domain_in)
    if domain:
        domain=domain.lower()
    else:
        log.debug("No eTLD for {}".format(domain))
    log.debug("Trimmed domain from {0} to {1}".format(domain_in,domain))
    return domain

@app.route("/stats", methods=['GET'])
def cache_info():
    res = cache.cache_info()
    res.update({ "rdap_good":memocache.get("rdap_good"), "rdap_fail": memocache.get("rdap_fail")})
    return jsonify(res)

@app.route("/stats-reset", methods=['GET'])
def cache_reset():
    cache.cache.stats(enable=True, reset=True)
    memocache.set("rdap_good",0)
    memocache.set("rdap_fail",0)
    return jsonify({"text":"Success"})

@app.route("/cache_browse", methods=['GET'])
def cache_browse():
    offset = request.args.get('offset',0, type=int)
    limit = request.args.get('limit',100, type=int)
    max_limit = min( limit, app.config.get("cache_browse_limit",100))
    resp = cache.cache_dump(offset,max_limit)
    return jsonify(resp)

@app.route("/cache_get", methods=['GET'])
def cache_get():
    domain = request.args.get('domain')
    if not domain:
        return jsonify({"text":"Try /cache_get?domain=markbaggett.com"})
    resp = cache.cache_get(domain)
    return jsonify(resp)

@app.route("/<string:domain>", methods=['GET'])
def get_domain(domain):
    log.debug("New Request for domain {0}.".format(domain) )
    #First try to get it from the Memory Cache
    domain = reduce_domain(domain)
    if not domain:
        return json_response("ERROR","ERROR","ERROR","ERROR",['No valid eTLD exists for this domain.'],'ERROR')                
    cache_data = cache.get(domain)
    log.debug("Is the domain in cache? {}".format(bool(cache_data)))
    if cache_data:
        return cache_data
    #If it isn't in the memory cache check the database
    else:
        freq_score = "disabled"
        alerts = []
        if app.config.get("enable_freq_scores"):
            freq_score = freq.probability(domain)
            if freq_score[0] < app.config.get("freq_avg_alert") and freq_score[1] < app.config.get("freq_word_alert"):
                alerts.append("LOW-FREQ-SCORES")
            elif freq_score[0] < app.config.get("freq_avg_alert") or freq_score[1] < app.config.get("freq_word_alert"):
                alerts.append("SUSPECT-FREQ-SCORE")
        if app.config.get("mode")=="rdap":
            alerts.append("YOUR-FIRST-CONTACT")
            rdap_seen_by_you = (datetime.datetime.utcnow()+datetime.timedelta(hours=app.config['timezone_offset']))
            rdap_seen_by_web, rdap_expires, rdap_error = rdap.get_domain_record(domain)
            if rdap_error:
                cache_expiration = app.config.get("rdap_error_ttl_days",7) * 86400
                alerts.append(rdap_error)
                resp = json_response("ERROR","ERROR","ERROR","ERROR",alerts,freq_score)
                if "YOUR-FIRST-CONTACT" in alerts:
                    alerts.remove("YOUR-FIRST-CONTACT")
                cache_resp = json_response("ERROR","ERROR","ERROR","ERROR",alerts, freq_score)
                cache.set(domain, cache_resp, seconds_to_live=cache_expiration)
                memocache.incr("rdap_fail")
                log.debug("rdap failed for domain {0}. Cached {1} seconds.".format(domain,cache_expiration) )
                return resp
            #if not expires and its doesn't expire for two years then its established.
            est_day_age = app.config.get("established_days_age",720)
            domain_age_seconds = (datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc) - rdap_seen_by_web).total_seconds()
            until_expires = (rdap_expires - datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc)).total_seconds()
            if domain_age_seconds > (est_day_age *86400):
                category = "ESTABLISHED"      
            else:
                category = "NEW"
                #It stays in the cache until it becomes established or it expires
                until_established = (est_day_age * 86400) - domain_age_seconds
                until_expires = min( until_expires, until_established )
            resp = json_response(rdap_seen_by_web, "RDAP", rdap_seen_by_you, category, alerts , freq_score )
            if "YOUR-FIRST-CONTACT" in alerts:
                alerts.remove("YOUR-FIRST-CONTACT")
            cache_response = json_response(rdap_seen_by_web, "RDAP", rdap_seen_by_you, category, alerts, freq_score )
            cache.set(domain, cache_response, until_expires)
            memocache.incr("rdap_good")
            log.debug("rdap success for domain {0}.Expires {1}.Cached {2} {3}".format(domain,rdap_expires,until_expires,category) )
            return resp 
        else:
            #Mode is not RDAP ask SANS ISC for domain data
            return "ISC Mode is still in development"
         

def config_app(working_path):
    working_path = pathlib.Path(working_path)
    if not working_path.is_dir():
        print("Sorry.  The path specified is invalid.  The directory you provide must already exist.")
        sys.exit(1)
    if sys.version_info.minor < 5 or sys.version_info.minor < 3:
        print("Please update your installation of Python.")
        sys.exit(1)
    #Initialize the global variables
    if not (working_path / "domain_stats.yaml").exists():
        print("No configuration file found.")
        sys.exit(1)
    yaml_path = str(working_path / "domain_stats.yaml")
    print("Using config {}".format( str(yaml_path)))
    app.config.update(config)
    os.chdir(working_path)
    return app

if __name__ == "__main__":
    x = config_app("/home/student/mydata")
    x.run(host="0.0.0.0",port=5730)




#!/usr/bin/env python3
from __future__ import division

import sys
import string
import re
import weakref
import json

from collections import Counter, defaultdict
from pprint import pprint

class node():
    """
    A class to represent a node.

    Note: Assigning a weight actually adds that value.  It doesn't set
    it to that value. For example node['c'] = 5    increases the value
    in node by 5

    Attributes
    ----------

    parent : node
        the parent to this node
    count : int
        the current count for the node
"""

    def __init__(self,parent):
        self.parent = parent
        self._pairs = Counter()
        self._cachecount = 0
        self._dirtycount = False

    def __getitem__(self,key):
        if self.parent.ignore_case and (key.islower() or key.isupper()):
            return  self._pairs[key.lower()] + self._pairs[key.upper()]
        else:
            return self._pairs[key]

    def __setitem__(self,key,value):

        # TODO: should consider using the __addr__ magic method instead
        # since the described behavior is this instead
        self._dirtycount = True
        self._pairs[key] += value

    @property
    def count(self):
        if self._dirtycount:
           self._cachecount = sum(self._pairs.values())
           self._dirtycount = False
        return self._cachecount

class FreqCounter(dict):
    """
    A class used for frequency counting.

    Attributes
    ----------

    ignore_case : bool
        TODO: Update
    ignorechars : str
        TODO: Update
    verbose : bool
        When enabled, print verbose messages to the console.
    count : int
        The number of entries in the table

    
    Methods
    -------

    toJSON()
        Returns a JSON representation of the Table as a string
    fromJSON(jsondata)
        Imports the JSON representation into the Table
    tally_str(line, weight=1)
        TODO: Update
    probability(line)
        Returns the probability of a given letter combination
    save(filename)
        Writes the table to a frequency file
    load(filename)
        Loads the table from a given frequency file
    printtable()
        Prints the JSON table to the console
    """
    def __init__(self, *args,**kwargs):
        self._table = defaultdict(lambda :node(self))
        self.ignore_case = False

        # TODO: consider refactoring to ignore_chars for consistency
        self.ignorechars = ""
        self.verbose = "verbose" in kwargs

    def __getitem__(self,key):
        return self._table[key]

    def __iter__(self):
        return iter(self._table)

    def __len__(self):
        return len(self._table)

    def toJSON(self):
        """
        Returns a string JSON reperesentation of the table.
        """
        serial = []
        for key,val in self._table.items():
            serial.append( (key, list(val._pairs.items())) )
        return json.dumps((self.ignore_case, self.ignorechars, serial))

    def fromJSON(self,jsondata):
        """
        Imports the table from a string JSON representation of the table

        Parameters
        ----------
        jsondata : str
            A string that represents the JSON Data

        Returns
        -------
        str
            the JSON representation of the table
        """

        # TODO: Raise an error if we get something that we didn't 
        # expect.
        args = json.loads(jsondata)
        if args:
            self.ignore_case = args[0]
            self.ignorechars = args[1]
            for outerkey,val in args[2]:
                self._table[outerkey] = node(self)
                for letter,count in val:
                   self._table[outerkey][letter] = count

    def tally_str(self,line,weight=1):
        """
        TODO: Update
        
        Parameters
        ----------
        line : string
            TODO: Update
        weight : int, optional
            the weight to be assigned to the pair (default = 1)
        """
        allpairs = re.findall(r"..", line)
        allpairs.extend(re.findall(r"..",line[1:]))
        for eachpair in allpairs:
            self[eachpair[0]][eachpair[1]] = weight

    def probability(self,line):
        """
        Calculates the probability of the word pair

        Parameters
        ----------
        line : str
            TODO: Update

        Returns
        -------
        float
            TODO: verify; the probability of the given word pair
        """
        allpairs = re.findall(r"..", line)
        allpairs.extend(re.findall(r"..",line[1:]))
        if self.verbose: 
            print("All pairs: {0}".format(allpairs))
        probs = []
        for eachpair in allpairs:
            pair = [eachpair[0], eachpair[1]]

            # check if any part of the pair should be ignored and alert
            # the user this was skipped
            if not all(x in self.ignorechars for x in pair):
                probs.append(self._probability(eachpair))
                if self.verbose: 
                    print ("Probability of {0}: {1}".format(eachpair,probs))
            elif self.verbose:
                print("Pair '{}' was ignored",format(self.ignorechars))
        if probs:
            average_probability = sum(probs) / len(probs) * 100
        else:
            average_probability = 0.0
        if self.verbose:
            print("Average Probability: {0}% \n\n".format(average_probability))
        
        totl1 = 0
        totl2 = 0
        for eachpair in allpairs:
            l1 = l2 = 0
            pair = [eachpair[0], eachpair[1]]

            if not all(x in self.ignorechars for x in pair):
                l1 += self[eachpair[0]].count
                if self.ignore_case and (eachpair[0].islower() or eachpair[0].isupper()):
                    l1 += self[eachpair[0].swapcase()].count
                l2 += self[eachpair[0]][eachpair[1]]
                if self.ignore_case and (eachpair[0].islower() or eachpair[0].isupper()):
                    l2 += self[eachpair[0].swapcase()][eachpair[1]]
                totl1 += l1
                totl2 += l2
                if self.verbose: 
                    print("Letter1:{0} Letter2:{1}  - This pair {2}:{3} {4}:{5}".format(
                        totl1,
                        totl2, 
                        eachpair[0],
                        l1,
                        eachpair[1],
                        l2
                    ))
        if (totl1 == 0) or (totl2 == 0):
            total_word_probability = 0.0
        else:
            total_word_probability = totl2/totl1 * 100
        if self.verbose: print("Total Word Probability: {0}/{1} = {2}".format(totl2, totl1, total_word_probability))
        return round(average_probability,4),round(total_word_probability,4)

    def _probability(self,twoletters):
        if self.ignore_case and (self[twoletters[0]].count == 0 and self[twoletters[0].swapcase()].count == 0):
            return 0.0
        if not self.ignore_case and self[twoletters[0]].count == 0:
            return 0.0
        if self.ignore_case and (twoletters[0].islower() or twoletters[0].isupper()):
            ignored_tot = sum([self[twoletters[0].lower()][eachlet] for eachlet in self.ignorechars]) + sum([self[twoletters[0].upper()][eachlet] for eachlet in self.ignorechars])
            let2 = self[twoletters[0].lower()][twoletters[1]] + self[twoletters[0].upper()][twoletters[1]]
            let1 = self[twoletters[0].lower()].count + self[twoletters[0].upper()].count
            if let1 - ignored_tot == 0:
                return 0.0
            return let2/(let1-ignored_tot)
        else:
            ignored_tot = sum([self[twoletters[0]][eachlet] for eachlet in self.ignorechars])
            if self[twoletters[0]].count - ignored_tot == 0:
                return 0.0
            return self[twoletters[0]][twoletters[1]] / (self[twoletters[0]].count - ignored_tot)

    def save(self,filename):
        try:
            file_handle =  open(filename, 'wb')
            file_handle.write(self.toJSON().encode("latin1"))
            file_handle.flush()
            file_handle.close()
        except Exception as e:
            print("Unable to write freq file :" + str(e))
            raise(e)

    def load(self,filename):
        try:
            file_handle =  open(filename,"rb")
            self.fromJSON(file_handle.read().decode("latin1"))
            file_handle.close()
        except Exception as e:
            print("Unable to load freq file :",str(e))
            raise(e)

    @property
    def count(self):
        return sum(map(lambda y:y.count, x._table.values()))

    def printtable(self):
        pprint(self.toJSON())



if __name__ == "__main__":
    import argparse
    import os
    parser=argparse.ArgumentParser()
    parser.add_argument(
        '-m',
        '--measure',
        required=False,
        help='Measure likelihood of a given string',
        dest='measure'
    )
    parser.add_argument(
        '-n',
        '--normal',
        required=False,
        help='Update the table based on the following normal string',
        dest='normal'
    )
    parser.add_argument(
        '-f',
        '--normalfile',
        required=False,
        help='Update the table based on the contents of the normal file',
        dest='normalfile'
    )
    parser.add_argument(
        '-p',
        '--print',
        action='store_true',
        required=False,
        help='Print a table of the most likely letters in order',
        dest='printtable'
    )
    parser.add_argument(
        '-c',
        '--create',
        action='store_true',
        required=False,
        help='Create a new empty frequency table',dest='create'
    )
    parser.add_argument(
        '-v',
        '--verbose',
        action='store_true',
        required=False,
        help='show calculation process',
        dest='verbose'
    )

    parser.add_argument(
        '-b',
        '--bulk',
        default=None,
        required=False,
        help='Measure every line in a file.',
        dest='bulk'
    )

    # TODO: break these up as well
    parser.add_argument('-t','--toggle_case_sensitivity',action='store_true',required=False,help='Ignore case in all future frequecy tabulations',dest='toggle_case')
    parser.add_argument('-s','--case_sensitive',action='store_true',required=False,help='Consider case in calculations. Default ignores case.',dest='case_sensitive')
    parser.add_argument('-w','--weight',type=int,default = 1, required=False,help='Affects weight of promote, update and update file (default is 1)',dest='weight')
    parser.add_argument('-e','--exclude',default = "\n\t~`!@#$%^&*()_+-", required=False,help='Provide a list of characters to ignore from the tabulations.',dest='exclude')
    parser.add_argument('freqtable',help='File storing character frequencies.')

    args=parser.parse_args()

    if args.verbose:
        fc = FreqCounter(verbose=True)
    else:
        fc = FreqCounter()
    if args.create and os.path.exists(args.freqtable):
        print("Frequency table already exists. " + args.freqtable)
        sys.exit(1)

    if not args.create:
        if not os.path.exists(args.freqtable):
           print("Frequency Character file not found. - %s " % (args.freqtable))
           raise(Exception("File not found."))
        fc.load(args.freqtable)

    if args.printtable: 
        fc.printtable()
    if args.normal: 
        fc.tally_str(args.normal, args.weight)
    if args.toggle_case:
        print("This feature has been depricated. By default all calculations ignore case.  Use -s to consider case.")
    fc.ignore_case = not args.case_sensitive
    fc.ignorechars = args.exclude
    if args.verbose: 
        print("Ignoring Case: {0}".format(fc.ignore_case))
    if args.verbose: 
        print("Ignoring Characters: {0}".format(fc.ignorechars))
    if args.normalfile:
        with open(args.normalfile,"rb") as filehandle:
            for eachline in filehandle:
                fc.tally_str(eachline.decode("latin1"))
    if args.measure: 
        print(fc.probability(args.measure))
    if args.bulk:
        with open(args.bulk,"rt") as filehandle:
            for eachline in filehandle:
                print(fc.probability(eachline),"-",eachline)
                
    fc.save(args.freqtable)